@@ -43,10 +43,18 @@
 #include "xlat/sigsys_codes.h"
 #include "xlat/sigtrap_codes.h"
 
+#ifdef ALPHA
+# include "xlat/alpha_trap_codes.h"
+#endif
+
 #ifdef SIGEMT
 # include "xlat/sigemt_codes.h"
 #endif
 
+#ifdef HAVE_SIGINFO_T_SI_PERF_FLAGS
+# include "xlat/sigtrap_perf_flags.h"
+#endif
+
 #ifndef SI_FROMUSER
 # define SI_FROMUSER(sip)	((sip)->si_code <= 0)
 #endif
@@ -82,7 +90,7 @@ print_si_code(const unsigned int si_code, const int si_signo)
 		case SIGCHLD:
 			code = xlookup(sigchld_codes, si_code);
 			break;
-		case SIGPOLL:
+		case SIGIO: /* SIGPOLL */
 			code = xlookup(sigpoll_codes, si_code);
 			break;
 		case SIGPROF:
@@ -139,6 +147,12 @@ print_si_info(struct tcb *tcp, const siginfo_t *sip)
 			printsigval(sip);
 			break;
 #endif
+		case SI_SIGIO:
+			tprint_struct_next();
+			PRINT_FIELD_D(*sip, si_band);
+			tprint_struct_next();
+			PRINT_FIELD_FD(*sip, si_fd, tcp);
+			break;
 		default:
 			printsigsource(tcp, sip);
 			if (sip->si_ptr)
@@ -157,33 +171,102 @@ print_si_info(struct tcb *tcp, const siginfo_t *sip)
 				PRINT_FIELD_OBJ_VAL(*sip, si_status, printsignal);
 			}
 			tprint_struct_next();
-			PRINT_FIELD_U(*sip, si_utime);
+			PRINT_FIELD_CLOCK_T(*sip, si_utime);
 			tprint_struct_next();
-			PRINT_FIELD_U(*sip, si_stime);
+			PRINT_FIELD_CLOCK_T(*sip, si_stime);
 			break;
 		case SIGILL:
+			tprint_struct_next();
+			PRINT_FIELD_PTR(*sip, si_addr);
+#if defined(SPARC) || defined(SPARC64)
+			tprint_struct_next();
+			PRINT_FIELD_D(*sip, si_trapno);
+#endif /* SPARC || SPARC64 */
+			break;
 		case SIGFPE:
+			tprint_struct_next();
+			PRINT_FIELD_PTR(*sip, si_addr);
+#ifdef ALPHA
+			tprint_struct_next();
+			PRINT_FIELD_XVAL_D(*sip, si_trapno, alpha_trap_codes,
+					   "GEN_???");
+#endif /* ALPHA */
+			break;
 		case SIGBUS:
 			tprint_struct_next();
 			PRINT_FIELD_PTR(*sip, si_addr);
+#if !defined(BUS_OPFETCH) && defined(HAVE_SIGINFO_T_SI_ADDR_LSB)
+			switch (sip->si_code) {
+			case BUS_MCEERR_AR:
+			case BUS_MCEERR_AO:
+				tprint_struct_next();
+				PRINT_FIELD_X(*sip, si_addr_lsb);
+				break;
+			}
+#endif /* !BUS_OPFETCH && HAVE_SIGINFO_T_SI_ADDR_LSB */
 			break;
 		case SIGSEGV:
 			tprint_struct_next();
 			PRINT_FIELD_PTR(*sip, si_addr);
-#ifdef HAVE_SIGINFO_T_SI_PKEY
+#if (!defined(SEGV_STACKFLOW) && defined(HAVE_SIGINFO_T_SI_LOWER)) \
+    || (!defined(__SEGV_PSTKOVF) && defined(HAVE_SIGINFO_T_SI_PKEY))
 			switch (sip->si_code) {
+# if !defined(SEGV_STACKFLOW) && defined(HAVE_SIGINFO_T_SI_LOWER)
+			case SEGV_BNDERR:
+				tprint_struct_next();
+				PRINT_FIELD_PTR(*sip, si_lower);
+				tprint_struct_next();
+				PRINT_FIELD_PTR(*sip, si_upper);
+				break;
+# endif /* !SEGV_STACKFLOW && HAVE_SIGINFO_T_SI_LOWER */
+# if !defined(__SEGV_PSTKOVF) && defined(HAVE_SIGINFO_T_SI_PKEY)
 			case SEGV_PKUERR:
 				tprint_struct_next();
 				PRINT_FIELD_U(*sip, si_pkey);
 				break;
+# endif /* !__SEGV_PSTKOVF && HAVE_SIGINFO_T_SI_PKEY */
 			}
-#endif
+#endif /* !SEGV_STACKFLOW && HAVE_SIGINFO_T_SI_LOWER
+	* || !__SEGV_PSTKOVF && HAVE_SIGINFO_T_SI_PKEY */
 			break;
-		case SIGPOLL:
+		case SIGTRAP:
+			tprint_struct_next();
+			PRINT_FIELD_PTR(*sip, si_addr);
+#if defined(ALPHA) || defined(HAVE_SIGINFO_T_SI_PERF_DATA)
+			switch (sip->si_code) {
+# ifdef ALPHA
+			case TRAP_UNK:
+				tprint_struct_next();
+				PRINT_FIELD_XVAL_D(*sip, si_trapno,
+						   alpha_trap_codes, "GEN_???");
+				break;
+# endif /* ALPHA */
+# ifdef HAVE_SIGINFO_T_SI_PERF_DATA
+			case TRAP_PERF:
+				tprint_struct_next();
+				PRINT_FIELD_X(*sip, si_perf_data);
+#  ifdef HAVE_SIGINFO_T_SI_PERF_TYPE
+				tprint_struct_next();
+				PRINT_FIELD_XVAL(*sip, si_perf_type,
+						 perf_type_id, "PERF_TYPE_???");
+#  endif /* HAVE_SIGINFO_T_SI_PERF_TYPE */
+#  ifdef HAVE_SIGINFO_T_SI_PERF_FLAGS
+				tprint_struct_next();
+				PRINT_FIELD_FLAGS(*sip, si_perf_flags,
+						  sigtrap_perf_flags,
+						  "TRAP_PERF_FLAG_???");
+#  endif /* HAVE_SIGINFO_T_SI_PERF_FLAGS */
+# endif /* HAVE_SIGINFO_T_SI_PERF_DATA */
+			}
+#endif /* ALPHA || HAVE_SIGINFO_T_SI_PERF_DATA */
+			break;
+		case SIGIO: /* SIGPOLL */
 			switch (sip->si_code) {
 			case POLL_IN: case POLL_OUT: case POLL_MSG:
 				tprint_struct_next();
 				PRINT_FIELD_D(*sip, si_band);
+				tprint_struct_next();
+				PRINT_FIELD_FD(*sip, si_fd, tcp);
 				break;
 			}
 			break;
