@@ -2,7 +2,7 @@
  * Check bpf syscall decoding.
  *
  * Copyright (c) 2015-2017 Dmitry V. Levin <ldv@strace.io>
- * Copyright (c) 2015-2021 The strace developers.
+ * Copyright (c) 2015-2022 The strace developers.
  * All rights reserved.
  *
  * SPDX-License-Identifier: GPL-2.0-or-later
@@ -26,15 +26,17 @@
 #include "print_fields.h"
 
 #include "xlat.h"
+#include "xlat/bpf_attach_type.h"
 #include "xlat/bpf_commands.h"
 #include "xlat/bpf_map_types.h"
 #include "xlat/bpf_prog_types.h"
+#include "xlat/bpf_test_run_flags.h"
 
 #if defined MPERS_IS_m32 || SIZEOF_KERNEL_LONG_T > 4
 # define BIG_ADDR(addr64_, addr32_) addr64_
 # define BIG_ADDR_MAYBE(addr_)
 #elif defined __arm__ || defined __i386__ || defined __mips__ \
-   || defined __powerpc__ || defined __riscv__ || defined __s390__ \
+   || defined __powerpc__ || defined __riscv || defined __s390__ \
    || defined __sparc__ || defined __tile__
 # define BIG_ADDR(addr64_, addr32_) addr64_ " or " addr32_
 # define BIG_ADDR_MAYBE(addr_) addr_ " or "
@@ -85,16 +87,18 @@ union bpf_attr_data {
 	BPF_ATTR_DATA_FIELD(BPF_MAP_DELETE_BATCH);
 	BPF_ATTR_DATA_FIELD(BPF_LINK_CREATE);
 	BPF_ATTR_DATA_FIELD(BPF_LINK_UPDATE);
+	BPF_ATTR_DATA_FIELD(BPF_LINK_GET_FD_BY_ID);
 	char char_data[256];
 };
 
 struct bpf_attr_check {
 	union bpf_attr_data data;
 	size_t size;
+	size_t iters;
 	const char *str;
-	void (*init_fn)(struct bpf_attr_check *check);
+	void (*init_fn)(struct bpf_attr_check *check, size_t idx);
 	void (*print_fn)(const struct bpf_attr_check *check,
-			 unsigned long addr);
+			 unsigned long addr, size_t idx);
 };
 
 struct bpf_check {
@@ -106,7 +110,7 @@ struct bpf_check {
 
 static const kernel_ulong_t long_bits = (kernel_ulong_t) 0xfacefeed00000000ULL;
 static const char *errstr;
-static unsigned int sizeof_attr = sizeof(union bpf_attr_data);
+static const unsigned int sizeof_attr = sizeof(union bpf_attr_data);
 static unsigned int page_size;
 static unsigned long end_of_page;
 
@@ -143,10 +147,11 @@ sys_bpf(kernel_ulong_t cmd, kernel_ulong_t attr, kernel_ulong_t size)
 #endif
 
 static void
-print_bpf_attr(const struct bpf_attr_check *check, unsigned long addr)
+print_bpf_attr(const struct bpf_attr_check *check, unsigned long addr,
+	       size_t idx)
 {
 	if (check->print_fn)
-		check->print_fn(check, addr);
+		check->print_fn(check, addr, idx);
 	else
 		printf("%s", check->str);
 }
@@ -157,6 +162,7 @@ test_bpf(const struct bpf_check *cmd_check)
 	const struct bpf_attr_check *check = 0;
 	const union bpf_attr_data *data = 0;
 	unsigned int offset = 0;
+	size_t j = 0;
 
 	/* zero addr */
 	sys_bpf(cmd_check->cmd, 0, long_bits | sizeof(union bpf_attr_data));
@@ -171,26 +177,30 @@ test_bpf(const struct bpf_check *cmd_check)
 
 	for (size_t i = 0; i < cmd_check->count; i++) {
 		check = &cmd_check->checks[i];
-		if (check->init_fn)
-			check->init_fn((struct bpf_attr_check *) check);
-		data = &check->data;
-		offset = check->size;
+		for (j = 0; j < MAX(check->iters, 1); j++) {
+			if (check->init_fn)
+				check->init_fn((struct bpf_attr_check *) check, j);
+			data = &check->data;
+			offset = check->size;
 
-		addr = end_of_page - offset;
-		memcpy((void *) addr, data, offset);
+			addr = end_of_page - offset;
+			memcpy((void *) addr, data, offset);
 
-		/* starting piece of bpf_attr_data */
-		sys_bpf(cmd_check->cmd, addr, offset);
-		printf("bpf(%s, {", cmd_check->cmd_str);
-		print_bpf_attr(check, addr);
-		printf("}, %u) = %s\n", offset, errstr);
+			/* starting piece of bpf_attr_data */
+			sys_bpf(cmd_check->cmd, addr, offset);
+			printf("bpf(%s, {", cmd_check->cmd_str);
+			print_bpf_attr(check, addr, j);
+			printf("}, %u) = %s\n", offset, errstr);
 
-		/* short read of the starting piece */
-		sys_bpf(cmd_check->cmd, addr + 1, offset);
-		printf("bpf(%s, %#lx, %u) = %s\n",
-		       cmd_check->cmd_str, addr + 1, offset, errstr);
+			/* short read of the starting piece */
+			sys_bpf(cmd_check->cmd, addr + 1, offset);
+			printf("bpf(%s, %#lx, %u) = %s\n",
+			       cmd_check->cmd_str, addr + 1, offset, errstr);
+		}
 	}
 
+	j = MAX(check->iters, 1) - 1;
+
 	if (offset < sizeof_attr) {
 		/* short read of the whole bpf_attr_data */
 		memcpy((void *) end_of_page - sizeof_attr + 1, data, offset);
@@ -206,7 +216,7 @@ test_bpf(const struct bpf_check *cmd_check)
 		memset((void *) addr + offset, 0, sizeof_attr - offset);
 		sys_bpf(cmd_check->cmd, addr, sizeof_attr);
 		printf("bpf(%s, {", cmd_check->cmd_str);
-		print_bpf_attr(check, addr);
+		print_bpf_attr(check, addr, j);
 		printf("}, %u) = %s\n", sizeof_attr, errstr);
 
 		/* non-zero bytes after the relevant part */
@@ -214,7 +224,7 @@ test_bpf(const struct bpf_check *cmd_check)
 			       sizeof_attr - offset, '0', 10);
 		sys_bpf(cmd_check->cmd, addr, sizeof_attr);
 		printf("bpf(%s, {", cmd_check->cmd_str);
-		print_bpf_attr(check, addr);
+		print_bpf_attr(check, addr, j);
 		printf(", ");
 		print_extra_data((char *) addr, offset,
 				 sizeof_attr - offset);
@@ -235,7 +245,7 @@ test_bpf(const struct bpf_check *cmd_check)
 	memset((void *) addr + offset, 0, page_size - offset);
 	sys_bpf(cmd_check->cmd, addr, page_size);
 	printf("bpf(%s, {", cmd_check->cmd_str);
-	print_bpf_attr(check, addr);
+	print_bpf_attr(check, addr, j);
 	printf("}, %u) = %s\n", page_size, errstr);
 
 	/* non-zero bytes after the whole bpf_attr_data */
@@ -243,7 +253,7 @@ test_bpf(const struct bpf_check *cmd_check)
 		       page_size - offset, '0', 10);
 	sys_bpf(cmd_check->cmd, addr, page_size);
 	printf("bpf(%s, {", cmd_check->cmd_str);
-	print_bpf_attr(check, addr);
+	print_bpf_attr(check, addr, j);
 	printf(", ");
 	print_extra_data((char *) addr, offset,
 			 page_size - offset);
@@ -256,13 +266,13 @@ test_bpf(const struct bpf_check *cmd_check)
 }
 
 static void
-init_BPF_MAP_CREATE_attr7(struct bpf_attr_check *check)
+init_BPF_MAP_CREATE_attr7(struct bpf_attr_check *check, size_t idx)
 {
 	struct BPF_MAP_CREATE_struct *attr = &check->data.BPF_MAP_CREATE_data;
 	attr->map_ifindex = ifindex_lo();
 }
 
-static_assert(ARRAY_SIZE(bpf_map_types_xdata) == 30,
+static_assert(ARRAY_SIZE(bpf_map_types_xdata) == 31,
 	      "The map_type for tests 1 and 2 below needs to be updated");
 static struct bpf_attr_check BPF_MAP_CREATE_checks[] = {
 	{
@@ -295,7 +305,7 @@ static struct bpf_attr_check BPF_MAP_CREATE_checks[] = {
 	},
 	{ /* 2 */
 		.data = { .BPF_MAP_CREATE_data = {
-			.map_type = 30,
+			.map_type = 31,
 			.key_size = 0xface1e55,
 			.value_size = 0xbadc0ded,
 			.max_entries = 0xbeefcafe,
@@ -306,7 +316,7 @@ static struct bpf_attr_check BPF_MAP_CREATE_checks[] = {
 			.map_ifindex = 3141592653,
 		} },
 		.size = offsetofend(struct BPF_MAP_CREATE_struct, map_ifindex),
-		.str = "map_type=0x1e /* BPF_MAP_TYPE_??? */"
+		.str = "map_type=0x1f /* BPF_MAP_TYPE_??? */"
 		       ", key_size=4207812181, value_size=3134983661"
 		       ", max_entries=3203386110"
 		       ", map_flags=0xffffe000 /* BPF_F_??? */"
@@ -469,6 +479,26 @@ static struct bpf_attr_check BPF_MAP_CREATE_checks[] = {
 		       ", btf_value_type_id=3405705229"
 		       ", btf_vmlinux_value_type_id=3735929054"
 	},
+	{ /* 10 */
+		.data = { .BPF_MAP_CREATE_data = {
+			.map_type = BPF_MAP_TYPE_BLOOM_FILTER,
+			.map_extra = 4
+		} },
+		.size = offsetofend(struct BPF_MAP_CREATE_struct, map_extra),
+		.str = "map_type=BPF_MAP_TYPE_BLOOM_FILTER"
+		       ", key_size=0"
+		       ", value_size=0"
+		       ", max_entries=0"
+		       ", map_flags=0"
+		       ", inner_map_fd=0" FD0_PATH
+		       ", map_name=\"\""
+		       ", map_ifindex=0"
+		       ", btf_fd=0" FD0_PATH
+		       ", btf_key_type_id=0"
+		       ", btf_value_type_id=0"
+		       ", btf_vmlinux_value_type_id=0"
+		       ", map_extra=4"
+	},
 };
 
 static const struct bpf_attr_check BPF_MAP_LOOKUP_ELEM_checks[] = {
@@ -596,7 +626,7 @@ get_log_buf_tail(void)
 #endif
 
 static void
-init_BPF_PROG_LOAD_attr3(struct bpf_attr_check *check)
+init_BPF_PROG_LOAD_attr3(struct bpf_attr_check *check, size_t idx)
 {
 	struct BPF_PROG_LOAD_struct *attr = &check->data.BPF_PROG_LOAD_data;
 
@@ -606,7 +636,8 @@ init_BPF_PROG_LOAD_attr3(struct bpf_attr_check *check)
 }
 
 static void
-print_BPF_PROG_LOAD_attr3(const struct bpf_attr_check *check, unsigned long addr)
+print_BPF_PROG_LOAD_attr3(const struct bpf_attr_check *check,
+			  unsigned long addr, size_t idx)
 {
 	printf("prog_type=BPF_PROG_TYPE_SOCKET_FILTER, insn_cnt=%u"
 	       ", insns=" INSNS_FMT ", license=\"%s\", log_level=2718281828"
@@ -619,7 +650,7 @@ print_BPF_PROG_LOAD_attr3(const struct bpf_attr_check *check, unsigned long addr
 }
 
 static void
-init_BPF_PROG_LOAD_attr4(struct bpf_attr_check *check)
+init_BPF_PROG_LOAD_attr4(struct bpf_attr_check *check, size_t idx)
 {
 	struct BPF_PROG_LOAD_struct *attr = &check->data.BPF_PROG_LOAD_data;
 
@@ -632,7 +663,8 @@ init_BPF_PROG_LOAD_attr4(struct bpf_attr_check *check)
 }
 
 static void
-print_BPF_PROG_LOAD_attr4(const struct bpf_attr_check *check, unsigned long addr)
+print_BPF_PROG_LOAD_attr4(const struct bpf_attr_check *check,
+			  unsigned long addr, size_t idx)
 {
 	printf("prog_type=BPF_PROG_TYPE_UNSPEC, insn_cnt=%u, insns=" INSNS_FMT
 	       ", license=\"%s\", log_level=2718281828, log_size=4"
@@ -766,9 +798,9 @@ static struct bpf_attr_check BPF_PROG_LOAD_checks[] = {
 			.line_info_cnt = 0xdad7bef8,
 			.attach_btf_id = 0xdad7befa,
 			.attach_prog_fd = 0xbadc0def,
+			.fd_array = 0xfaceb00c,
 		} },
-		.size = offsetofend(struct BPF_PROG_LOAD_struct,
-				    attach_prog_fd),
+		.size = offsetofend(struct BPF_PROG_LOAD_struct, fd_array),
 		.str = "prog_type=BPF_PROG_TYPE_UNSPEC"
 		       ", insn_cnt=0"
 		       ", insns=NULL"
@@ -790,11 +822,12 @@ static struct bpf_attr_check BPF_PROG_LOAD_checks[] = {
 		       ", line_info_cnt=3671572216"
 		       ", attach_btf_id=3671572218"
 		       ", attach_prog_fd=-1159983633"
+		       ", fd_array=0xfaceb00c"
 	},
 };
 
 static void
-init_BPF_OBJ_PIN_attr(struct bpf_attr_check *check)
+init_BPF_OBJ_PIN_attr(struct bpf_attr_check *check, size_t idx)
 {
 	struct BPF_OBJ_PIN_struct *attr = &check->data.BPF_OBJ_PIN_data;
 	attr->pathname = (uintptr_t) pathname;
@@ -913,6 +946,30 @@ static const struct bpf_attr_check BPF_PROG_TEST_RUN_checks[] = {
 		       ", data_out=0xfacef33dbadc4ded"
 		       ", repeat=4207410904, duration=4207541978}"
 	},
+	{
+		.data = { .BPF_PROG_TEST_RUN_data = {
+			.prog_fd = -1,
+			.retval = 0xfac1fed2,
+			.data_size_in = 0xfac3fed4,
+			.data_size_out = 0xfac5fed6,
+			.data_in = (uint64_t) 0xfacef11dbadc2dedULL,
+			.data_out = (uint64_t) 0xfacef33dbadc4dedULL,
+			.repeat = 0xfac7fed8,
+			.duration = 0xfac9feda,
+			.ctx_size_in = 0,
+			.ctx_size_out = 0xfacdfede,
+			.ctx_in = (uint64_t) 0xfacef55dbadc6dedULL,
+		} },
+		.size = offsetofend(struct BPF_PROG_TEST_RUN_struct, ctx_in),
+		.str = "test={prog_fd=-1, retval=4207017682"
+		       ", data_size_in=4207148756, data_size_out=4207279830"
+		       ", data_in=0xfacef11dbadc2ded"
+		       ", data_out=0xfacef33dbadc4ded"
+		       ", repeat=4207410904"
+		       ", duration=4207541978"
+		       ", ctx_size_in=0, ctx_size_out=4207804126"
+		       ", ctx_in=0xfacef55dbadc6ded, ctx_out=NULL}"
+	},
 	{
 		.data = { .BPF_PROG_TEST_RUN_data = {
 			.prog_fd = -1,
@@ -926,9 +983,11 @@ static const struct bpf_attr_check BPF_PROG_TEST_RUN_checks[] = {
 			.ctx_size_in = 0xfacbfedc,
 			.ctx_size_out = 0xfacdfede,
 			.ctx_in = (uint64_t) 0xfacef55dbadc6dedULL,
-			.ctx_out = (uint64_t) 0xfacef77dbadc8dedULL
+			.ctx_out = (uint64_t) 0xfacef77dbadc8dedULL,
+			.flags = BPF_F_TEST_RUN_ON_CPU|BPF_F_TEST_XDP_LIVE_FRAMES,
+			.cpu = 0,
 		} },
-		.size = offsetofend(struct BPF_PROG_TEST_RUN_struct, ctx_out),
+		.size = offsetofend(struct BPF_PROG_TEST_RUN_struct, cpu),
 		.str = "test={prog_fd=-1, retval=4207017682"
 		       ", data_size_in=4207148756, data_size_out=4207279830"
 		       ", data_in=0xfacef11dbadc2ded"
@@ -938,8 +997,40 @@ static const struct bpf_attr_check BPF_PROG_TEST_RUN_checks[] = {
 		       ", ctx_size_in=4207673052"
 		       ", ctx_size_out=4207804126"
 		       ", ctx_in=0xfacef55dbadc6ded"
-		       ", ctx_out=0xfacef77dbadc8ded}"
-	}
+		       ", ctx_out=0xfacef77dbadc8ded"
+		       ", flags=BPF_F_TEST_RUN_ON_CPU|BPF_F_TEST_XDP_LIVE_FRAMES"
+		       ", cpu=0}"
+	},
+	{
+		.data = { .BPF_PROG_TEST_RUN_data = {
+			.prog_fd = -1,
+			.retval = 0xfac1fed2,
+			.data_size_in = 0xfac3fed4,
+			.data_size_out = 0xfac5fed6,
+			.data_in = (uint64_t) 0xfacef11dbadc2dedULL,
+			.data_out = (uint64_t) 0xfacef33dbadc4dedULL,
+			.repeat = 0xfac7fed8,
+			.duration = 0xfac9feda,
+			.ctx_size_in = 0,
+			.ctx_size_out = 0,
+			.ctx_in = 0,
+			.ctx_out = 0,
+			.flags = 0xfffffffc,
+			.cpu = 3141592653,
+			.batch_size = 2718281828,
+		} },
+		.size = offsetofend(struct BPF_PROG_TEST_RUN_struct, batch_size),
+		.str = "test={prog_fd=-1, retval=4207017682"
+		       ", data_size_in=4207148756, data_size_out=4207279830"
+		       ", data_in=0xfacef11dbadc2ded"
+		       ", data_out=0xfacef33dbadc4ded"
+		       ", repeat=4207410904"
+		       ", duration=4207541978"
+		       ", ctx_size_in=0, ctx_size_out=0"
+		       ", ctx_in=NULL, ctx_out=NULL"
+		       ", flags=0xfffffffc /* BPF_F_??? */"
+		       ", cpu=3141592653, batch_size=2718281828}"
+	},
 };
 
 static const struct bpf_attr_check BPF_PROG_GET_NEXT_ID_checks[] = {
@@ -980,6 +1071,7 @@ static const struct bpf_attr_check BPF_PROG_GET_NEXT_ID_checks[] = {
 
 #define BPF_MAP_GET_NEXT_ID_checks BPF_PROG_GET_NEXT_ID_checks
 #define BPF_BTF_GET_NEXT_ID_checks BPF_PROG_GET_NEXT_ID_checks
+#define BPF_LINK_GET_NEXT_ID_checks BPF_PROG_GET_NEXT_ID_checks
 
 static const struct bpf_attr_check BPF_PROG_GET_FD_BY_ID_checks[] = {
 	{
@@ -1060,7 +1152,7 @@ static uint32_t prog_load_ids[] = { 0, 1, 0xffffffff, 2718281828, };
 uint32_t *prog_load_ids_ptr;
 
 static void
-init_BPF_PROG_QUERY_attr4(struct bpf_attr_check *check)
+init_BPF_PROG_QUERY_attr4(struct bpf_attr_check *check, size_t idx)
 {
 	struct BPF_PROG_QUERY_struct *attr = &check->data.BPF_PROG_QUERY_data;
 
@@ -1073,7 +1165,8 @@ init_BPF_PROG_QUERY_attr4(struct bpf_attr_check *check)
 }
 
 static void
-print_BPF_PROG_QUERY_attr4(const struct bpf_attr_check *check, unsigned long addr)
+print_BPF_PROG_QUERY_attr4(const struct bpf_attr_check *check,
+			   unsigned long addr, size_t idx)
 {
 	printf("query={target_fd=-1153374643"
 	       ", attach_type=0xfeedface /* BPF_??? */"
@@ -1088,7 +1181,7 @@ print_BPF_PROG_QUERY_attr4(const struct bpf_attr_check *check, unsigned long add
 }
 
 static void
-init_BPF_PROG_QUERY_attr5(struct bpf_attr_check *check)
+init_BPF_PROG_QUERY_attr5(struct bpf_attr_check *check, size_t idx)
 {
 	struct BPF_PROG_QUERY_struct *attr = &check->data.BPF_PROG_QUERY_data;
 
@@ -1101,7 +1194,8 @@ init_BPF_PROG_QUERY_attr5(struct bpf_attr_check *check)
 }
 
 static void
-print_BPF_PROG_QUERY_attr5(const struct bpf_attr_check *check, unsigned long addr)
+print_BPF_PROG_QUERY_attr5(const struct bpf_attr_check *check,
+			   unsigned long addr, size_t idx)
 {
 	printf("query={target_fd=-1153374643"
 	       ", attach_type=0xfeedface /* BPF_??? */"
@@ -1128,13 +1222,13 @@ static struct bpf_attr_check BPF_PROG_QUERY_checks[] = {
 	{ /* 1 */
 		.data = { .BPF_PROG_QUERY_data = {
 			.target_fd = 3141592653U,
-			.attach_type = 40,
+			.attach_type = 42,
 			.query_flags = 1,
 			.attach_flags = 3,
 		} },
 		.size = offsetofend(struct BPF_PROG_QUERY_struct, attach_flags),
 		.str = "query={target_fd=-1153374643"
-		       ", attach_type=BPF_SK_REUSEPORT_SELECT_OR_MIGRATE"
+		       ", attach_type=BPF_TRACE_KPROBE_MULTI"
 		       ", query_flags=BPF_F_QUERY_EFFECTIVE"
 		       ", attach_flags=BPF_F_ALLOW_OVERRIDE|BPF_F_ALLOW_MULTI"
 		       ", prog_ids=NULL, prog_cnt=0}",
@@ -1142,7 +1236,7 @@ static struct bpf_attr_check BPF_PROG_QUERY_checks[] = {
 	{ /* 2 */
 		.data = { .BPF_PROG_QUERY_data = {
 			.target_fd = 3141592653U,
-			.attach_type = 41,
+			.attach_type = 43,
 			.query_flags = 0xfffffffe,
 			.attach_flags = 0xfffffff8,
 			.prog_ids = 0xffffffffffffffffULL,
@@ -1150,7 +1244,7 @@ static struct bpf_attr_check BPF_PROG_QUERY_checks[] = {
 		} },
 		.size = offsetofend(struct BPF_PROG_QUERY_struct, prog_cnt),
 		.str = "query={target_fd=-1153374643"
-		       ", attach_type=0x29 /* BPF_??? */"
+		       ", attach_type=0x2b /* BPF_??? */"
 		       ", query_flags=0xfffffffe /* BPF_F_QUERY_??? */"
 		       ", attach_flags=0xfffffff8 /* BPF_F_??? */"
 		       ", prog_ids="
@@ -1200,7 +1294,7 @@ static struct bpf_attr_check BPF_PROG_QUERY_checks[] = {
 
 
 static void
-init_BPF_RAW_TRACEPOINT_attr2(struct bpf_attr_check *check)
+init_BPF_RAW_TRACEPOINT_attr2(struct bpf_attr_check *check, size_t idx)
 {
 	/* TODO: test the 128 byte limit */
 	static const char tp_name[] = "0123456789qwertyuiop0123456789qwe";
@@ -1243,7 +1337,7 @@ static struct bpf_attr_check BPF_RAW_TRACEPOINT_OPEN_checks[] = {
 };
 
 static void
-init_BPF_BTF_LOAD_attr(struct bpf_attr_check *check)
+init_BPF_BTF_LOAD_attr(struct bpf_attr_check *check, size_t idx)
 {
 	static const char sample_btf_data[] = "bPf\0daTum";
 
@@ -1391,24 +1485,314 @@ static const struct bpf_attr_check BPF_MAP_DELETE_BATCH_checks[] = {
 	}
 };
 
-static const struct bpf_attr_check BPF_LINK_CREATE_checks[] = {
-	{
-		.data = { .BPF_LINK_CREATE_data = { .prog_fd = -1, .target_fd = -2 } },
-		.size = offsetofend(struct BPF_LINK_CREATE_struct, flags),
-		.str = "link_create={prog_fd=-1, target_fd=-2"
+static void
+init_BPF_LINK_CREATE_attr1(struct bpf_attr_check *check, size_t idx)
+{
+	struct BPF_LINK_CREATE_struct *attr = &check->data.BPF_LINK_CREATE_data;
+
+	attr->attach_type = idx;
+}
+
+static void
+print_BPF_LINK_CREATE_attr1(const struct bpf_attr_check *check,
+			    unsigned long addr, size_t idx)
+{
+	printf("link_create={prog_fd=-1, target_fd=-559038737"
+	       ", attach_type=%s, flags=0x4}",
+	       sprintxval(bpf_attach_type, idx, "BPF_???"));
+}
+
+/* Keep sorted */
+static const uint8_t special_attach_types[] =
+	{ 0, BPF_TRACE_ITER, BPF_PERF_EVENT, BPF_TRACE_KPROBE_MULTI };
+
+static void
+init_BPF_LINK_CREATE_attr2(struct bpf_attr_check *check, size_t idx)
+{
+	struct BPF_LINK_CREATE_struct *attr = &check->data.BPF_LINK_CREATE_data;
+
+	/* skip special_attach_types */
+	for (size_t i = 0; i < ARRAY_SIZE(special_attach_types)
+			   && idx >= special_attach_types[i]; i++, idx++);
+
+	attr->attach_type = idx;
+
+	check->data.char_data[19] = ' ';
+	check->data.char_data[23] = 'O';
+	check->data.char_data[27] = 'H';
+	check->data.char_data[31] = ' ';
+	check->data.char_data[35] = 'H';
+	check->data.char_data[39] = 'A';
+	check->data.char_data[43] = 'I';
+	check->data.char_data[47] = '!';
+}
+
+static void
+print_BPF_LINK_CREATE_attr2(const struct bpf_attr_check *check,
+			    unsigned long addr, size_t idx)
+{
+	/* skip special_attach_types */
+	for (size_t i = 0; i < ARRAY_SIZE(special_attach_types)
+			   && idx >= special_attach_types[i]; i++, idx++);
+
+	printf("link_create={prog_fd=-1, target_fd=-559038737"
+	       ", attach_type=%s, flags=0xbadc0ded}, "
+#if VERBOSE
+	       "extra_data=\"\\x00\\x00\\x00\\x20\\x00\\x00\\x00\\x4f"
+	       "\\x00\\x00\\x00\\x48\\x00\\x00\\x00\\x20\\x00\\x00\\x00\\x48"
+	       "\\x00\\x00\\x00\\x41\\x00\\x00\\x00\\x49\\x00\\x00\\x00\\x21\""
+	       " /* bytes 16..47 */"
+#else
+	       "..."
+#endif
+	       ,
+	       sprintxval(bpf_attach_type, idx, "BPF_???"));
+}
+
+static const int iter_info_data[] = { 0, 42, 314159265, 0xbadc0ded, -1 };
+static int *iter_info_data_p;
+
+static void
+init_BPF_LINK_CREATE_attr7(struct bpf_attr_check *check, size_t idx)
+{
+	struct BPF_LINK_CREATE_struct *attr = &check->data.BPF_LINK_CREATE_data;
+
+	close(iter_info_data[1]);
+
+	if (!iter_info_data_p) {
+		iter_info_data_p = tail_memdup(iter_info_data,
+					       sizeof(iter_info_data));
+	}
+
+	attr->iter_info = (uintptr_t) iter_info_data_p;
+	attr->iter_info_len = ARRAY_SIZE(iter_info_data) + idx;
+}
+
+static void
+print_BPF_LINK_CREATE_attr7(const struct bpf_attr_check *check,
+			    unsigned long addr, size_t idx)
+{
+	printf("link_create={prog_fd=0" FD0_PATH ", target_fd=0" FD0_PATH
+	       ", attach_type=BPF_TRACE_ITER, flags=0"
+	       ", iter_info=[{map={map_fd=0" FD0_PATH "}}, {map={map_fd=42}}"
+	       ", {map={map_fd=314159265}}, {map={map_fd=-1159983635}}"
+	       ", {map={map_fd=-1}}");
+	if (idx) {
+		printf(", ... /* %p */",
+		       iter_info_data_p + ARRAY_SIZE(iter_info_data));
+	}
+	printf("], iter_info_len=%zu}", ARRAY_SIZE(iter_info_data) + idx);
+
+}
+
+static const char *syms_data[] = { "foo", NULL, "OH\0HAI",
+				   "abcdefghijklmnopqrstuvwxyz0123456789" };
+static char **syms_data_p;
+static const uint64_t addrs_data[] = { 0, 1, 0xbadc0ded,
+				       0xfacefeeddeadc0deULL };
+static uint64_t *addrs_data_p;
+
+static_assert(ARRAY_SIZE(syms_data) == ARRAY_SIZE(addrs_data),
+	      "syms_data and addrs_data have to have the same element count");
+
+static void
+init_BPF_LINK_CREATE_attr12(struct bpf_attr_check *check, size_t idx)
+{
+	struct BPF_LINK_CREATE_struct *attr = &check->data.BPF_LINK_CREATE_data;
+
+	if (!syms_data_p)
+		syms_data_p = tail_memdup(syms_data, sizeof(syms_data));
+	if (!addrs_data_p)
+		addrs_data_p = tail_memdup(addrs_data, sizeof(addrs_data));
+
+	attr->kprobe_multi.cnt = ARRAY_SIZE(syms_data) + idx;
+	attr->kprobe_multi.syms = (uintptr_t) syms_data_p;
+	attr->kprobe_multi.addrs = (uintptr_t) addrs_data_p;
+	attr->kprobe_multi.cookies = (uintptr_t) addrs_data_p;
+}
+
+static void
+print_BPF_LINK_CREATE_attr12(const struct bpf_attr_check *check,
+			     unsigned long addr, size_t idx)
+{
+	printf("link_create={prog_fd=0" FD0_PATH ", target_fd=0" FD0_PATH
+	       ", attach_type=BPF_TRACE_KPROBE_MULTI, flags=0"
+	       ", kprobe_multi={flags=BPF_F_KPROBE_MULTI_RETURN|0xfacebeee"
+	       ", cnt=%zu", ARRAY_SIZE(syms_data) + idx);
+	printf(", syms=[\"foo\", NULL, \"OH\""
+	       ", \"abcdefghijklmnopqrstuvwxyz012345\"...");
+	if (idx)
+		printf(", ... /* %p */", syms_data_p + ARRAY_SIZE(syms_data));
+	for (size_t i = 0; i < 2; i++) {
+		printf("], %s=[0, 0x1, 0xbadc0ded, 0xfacefeeddeadc0de",
+		       i ? "cookies" : "addrs");
+		if (idx) {
+			printf(", ... /* %p */",
+			       addrs_data_p + ARRAY_SIZE(addrs_data));
+		}
+	}
+	printf("]}}");
+}
+
+static struct bpf_attr_check BPF_LINK_CREATE_checks[] = {
+	{ /* 0 */
+		.data = { .BPF_LINK_CREATE_data = { .prog_fd = 0, .target_fd = 0 } },
+		.size = offsetofend(struct BPF_LINK_CREATE_struct, target_fd),
+		.str = "link_create={prog_fd=0" FD0_PATH ", target_fd=0" FD0_PATH
 		       ", attach_type=BPF_CGROUP_INET_INGRESS, flags=0}"
 	},
-	{
+	{ /* 1 */
+		.data = { .BPF_LINK_CREATE_data = {
+			.prog_fd = -1,
+			.target_fd = 0xdeadbeef,
+			.flags = 4
+		} },
+		.size = offsetofend(struct BPF_LINK_CREATE_struct, flags),
+		.iters = ARRAY_SIZE(bpf_attach_type_xdata),
+		.init_fn = init_BPF_LINK_CREATE_attr1,
+		.print_fn = print_BPF_LINK_CREATE_attr1,
+	},
+	{ /* 2 - all non-special attach_types */
 		.data = { .BPF_LINK_CREATE_data = {
 			.prog_fd = -1,
 			.target_fd = 0xdeadbeef,
 			.attach_type = 5,
-			.flags = 4
+			.flags = 0xbadc0ded
 		} },
-		.size = offsetofend(struct BPF_LINK_CREATE_struct, flags),
-		.str = "link_create={prog_fd=-1, target_fd=-559038737"
-		       ", attach_type=BPF_SK_SKB_STREAM_VERDICT, flags=0x4}"
-	}
+		.size = 48,
+		.iters = ARRAY_SIZE(bpf_attach_type_xdata)
+			 - ARRAY_SIZE(special_attach_types),
+		.init_fn = init_BPF_LINK_CREATE_attr2,
+		.print_fn = print_BPF_LINK_CREATE_attr2,
+	},
+
+	{ /* 3 */
+		.data = { .BPF_LINK_CREATE_data = {
+			.attach_type = 0,
+		} },
+		.size = offsetofend(struct BPF_LINK_CREATE_struct,
+				    target_btf_id),
+		.str = "link_create={prog_fd=0" FD0_PATH", target_fd=0" FD0_PATH
+		       ", attach_type=BPF_CGROUP_INET_INGRESS, flags=0}"
+	},
+	{ /* 4 */
+		.data = { .BPF_LINK_CREATE_data = {
+			.attach_type = 0,
+			.target_btf_id = 0xfacefeed,
+		} },
+		.size = offsetofend(struct BPF_LINK_CREATE_struct,
+				    target_btf_id),
+		.str = "link_create={prog_fd=0" FD0_PATH", target_fd=0" FD0_PATH
+		       ", attach_type=BPF_CGROUP_INET_INGRESS, flags=0"
+		       ", target_btf_id=4207869677}"
+	},
+
+	{ /* 5 */
+		.data = { .BPF_LINK_CREATE_data = {
+			.attach_type = 28,
+		} },
+		.size = offsetofend(struct BPF_LINK_CREATE_struct,
+				    iter_info_len),
+		.str = "link_create={prog_fd=0" FD0_PATH", target_fd=0" FD0_PATH
+		       ", attach_type=BPF_TRACE_ITER, flags=0"
+		       ", iter_info=NULL, iter_info_len=0}"
+	},
+	{ /* 6 */
+		.data = { .BPF_LINK_CREATE_data = {
+			.attach_type = 28,
+			.iter_info = 0xffffffff00000000,
+			.iter_info_len = 0xdeadface,
+		} },
+		.size = offsetofend(struct BPF_LINK_CREATE_struct,
+				    iter_info_len),
+		.str = "link_create={prog_fd=0" FD0_PATH", target_fd=0" FD0_PATH
+		       ", attach_type=BPF_TRACE_ITER, flags=0"
+		       ", iter_info=" BIG_ADDR("0xffffffff00000000", "NULL")
+		       ", iter_info_len=3735943886}"
+	},
+	{ /* 7 */
+		.data = { .BPF_LINK_CREATE_data = {
+			.attach_type = 28,
+		} },
+		.size = offsetofend(struct BPF_LINK_CREATE_struct,
+				    iter_info_len),
+		.iters = 2,
+		.init_fn = init_BPF_LINK_CREATE_attr7,
+		.print_fn = print_BPF_LINK_CREATE_attr7,
+	},
+
+	{ /* 8 */
+		.data = { .BPF_LINK_CREATE_data = {
+			.attach_type = 41,
+		} },
+		.size = offsetofend(struct BPF_LINK_CREATE_struct,
+				    perf_event.bpf_cookie),
+		.str = "link_create={prog_fd=0" FD0_PATH", target_fd=0" FD0_PATH
+		       ", attach_type=BPF_PERF_EVENT, flags=0"
+		       ", perf_event={bpf_cookie=0}}"
+	},
+	{ /* 9 */
+		.data = { .BPF_LINK_CREATE_data = {
+			.attach_type = 41,
+			.perf_event = { .bpf_cookie = 0xdeadc0defacecafeULL },
+		} },
+		.size = offsetofend(struct BPF_LINK_CREATE_struct,
+				    perf_event.bpf_cookie),
+		.str = "link_create={prog_fd=0" FD0_PATH", target_fd=0" FD0_PATH
+		       ", attach_type=BPF_PERF_EVENT, flags=0"
+		       ", perf_event={bpf_cookie=0xdeadc0defacecafe}}"
+	},
+
+	{ /* 10 */
+		.data = { .BPF_LINK_CREATE_data = {
+			.attach_type = 42,
+		} },
+		.size = offsetofend(struct BPF_LINK_CREATE_struct,
+				    kprobe_multi.cookies),
+		.str = "link_create={prog_fd=0" FD0_PATH", target_fd=0" FD0_PATH
+		       ", attach_type=BPF_TRACE_KPROBE_MULTI, flags=0"
+		       ", kprobe_multi={flags=0, cnt=0, syms=NULL, addrs=NULL"
+		       ", cookies=NULL}}"
+	},
+	{ /* 11 */
+		.data = { .BPF_LINK_CREATE_data = {
+			.attach_type = 42,
+			.kprobe_multi = {
+				.flags = 0xdeadc0de,
+				.cnt = 0xbadfaced,
+				.syms = 0xffffffff00000000,
+				.addrs = 0xffffffff00000000,
+				.cookies = 0xffffffff00000000,
+			},
+		} },
+		.size = offsetofend(struct BPF_LINK_CREATE_struct,
+				    kprobe_multi.cookies),
+		.str = "link_create={prog_fd=0" FD0_PATH", target_fd=0" FD0_PATH
+		       ", attach_type=BPF_TRACE_KPROBE_MULTI, flags=0"
+		       ", kprobe_multi={flags=0xdeadc0de /* BPF_F_??? */"
+		       ", cnt=3135220973"
+		       ", syms=" BIG_ADDR("0xffffffff00000000", "NULL")
+		       ", addrs=" BIG_ADDR("0xffffffff00000000", "NULL")
+		       ", cookies=" BIG_ADDR("0xffffffff00000000", "NULL") "}}"
+	},
+	/*
+	 * Note that here we rely on the fact that this attach_type has the
+	 * largest de-facto attr_size to get the additional checks performed
+	 * with the last check passed.
+	 */
+	{ /* 12 */
+		.data = { .BPF_LINK_CREATE_data = {
+			.attach_type = 42,
+			.kprobe_multi = {
+				.flags = 0xfacebeef,
+			}
+		} },
+		.size = offsetofend(struct BPF_LINK_CREATE_struct,
+				    kprobe_multi.cookies),
+		.iters = 2,
+		.init_fn = init_BPF_LINK_CREATE_attr12,
+		.print_fn = print_BPF_LINK_CREATE_attr12,
+	},
 };
 
 static const struct bpf_attr_check BPF_LINK_UPDATE_checks[] = {
@@ -1433,6 +1817,14 @@ static const struct bpf_attr_check BPF_LINK_UPDATE_checks[] = {
 	}
 };
 
+static const struct bpf_attr_check BPF_LINK_GET_FD_BY_ID_checks[] = {
+	{
+		.data = { .BPF_LINK_GET_FD_BY_ID_data = { .link_id = 0xdeadbeef } },
+		.size = offsetofend(struct BPF_LINK_GET_FD_BY_ID_struct, link_id),
+		.str = "link_id=3735928559"
+	}
+};
+
 
 #define CHK(cmd_) \
 	{ \
@@ -1474,6 +1866,8 @@ main(void)
 		CHK(BPF_MAP_DELETE_BATCH),
 		CHK(BPF_LINK_CREATE),
 		CHK(BPF_LINK_UPDATE),
+		CHK(BPF_LINK_GET_NEXT_ID),
+		CHK(BPF_LINK_GET_FD_BY_ID),
 	};
 
 	page_size = get_page_size();
