@@ -44,13 +44,6 @@
 
 #define ARR_ITEM(arr_, idx_) ((arr_)[(idx_) % ARRAY_SIZE(arr_)])
 
-/* Work around field name change in Linux commit v5.19-rc1~251^2~20. */
-#ifdef HAVE_STRUCT_IO_URING_RSRC_REGISTER_RESV
-# define RESV resv
-#else
-# define RESV flags
-#endif
-
 static const char path_null[] = "/dev/null";
 static const char path_full[] = "/dev/full";
 
@@ -176,7 +169,7 @@ main(void)
 
 
 	/* Invalid op */
-	static const unsigned int invalid_ops[] = { 0xbadc0dedU, 22 };
+	static const unsigned int invalid_ops[] = { 0xbadc0dedU, 24 };
 
 	for (size_t i = 0; i < ARRAY_SIZE(invalid_ops); i++) {
 		sys_io_uring_register(fd_null, invalid_ops[i], path_null,
@@ -347,12 +340,12 @@ main(void)
 	probe->ops[0].flags = 0;
 	probe->ops[0].resv2 = 0xbeefface;
 
-	probe->ops[1].op = 40;
+	probe->ops[1].op = 46;
 	probe->ops[1].resv = 0;
 	probe->ops[1].flags = IO_URING_OP_SUPPORTED;
 	probe->ops[1].resv2 = 0xdeadc0de;
 
-	probe->ops[2].op = 41;
+	probe->ops[2].op = 47;
 	probe->ops[2].resv = 0xaf;
 	probe->ops[2].flags = 0xbeef;
 	probe->ops[2].resv2 = 0;
@@ -367,32 +360,32 @@ main(void)
 	       ", ops_len=%hhu, resv2=[0, %#x, 0], ops=["
 	       "{op=" XLAT_FMT_U ", resv=0xde, flags=0, resv2=0xbeefface}, "
 	       "{op=" XLAT_FMT_U ", flags=" XLAT_FMT ", resv2=0xdeadc0de}, "
-	       "{op=41" NRAW(" /* IORING_OP_??? */") ", resv=0xaf, flags="
+	       "{op=47" NRAW(" /* IORING_OP_??? */") ", resv=0xaf, flags="
 	       XLAT_FMT "}, {op=254" NRAW(" /* IORING_OP_??? */")
 	       ", flags=0xc0de" NRAW(" /* IO_URING_OP_??? */") "}]}"
 #if RETVAL_INJECTED
 	       " => {last_op=" XLAT_FMT_U ", ops_len=%hhu, resv2=[0, %#x, 0], "
 	       "ops=[{op=" XLAT_FMT_U ", resv=0xde, flags=0, resv2=0xbeefface}"
 	       ", {op=" XLAT_FMT_U ", flags=" XLAT_FMT ", resv2=0xdeadc0de}"
-	       ", {op=41" NRAW(" /* IORING_OP_??? */") ", resv=0xaf, flags="
+	       ", {op=47" NRAW(" /* IORING_OP_??? */") ", resv=0xaf, flags="
 	       XLAT_FMT "}, {op=254" NRAW(" /* IORING_OP_??? */")
 	       ", flags=0xc0de" NRAW(" /* IO_URING_OP_??? */") "}, ...]}"
 #endif
 	       ", 4) = %s\n",
 	       fd_null, path_null, XLAT_ARGS(IORING_REGISTER_PROBE),
 	       XLAT_ARGS(IORING_OP_EPOLL_CTL), probe->ops_len, probe->resv2[1],
-	       XLAT_ARGS(IORING_OP_NOP), XLAT_ARGS(IORING_OP_MSG_RING),
+	       XLAT_ARGS(IORING_OP_NOP), XLAT_ARGS(IORING_OP_URING_CMD),
 	       XLAT_ARGS(IO_URING_OP_SUPPORTED),
 	       XLAT_ARGS(IO_URING_OP_SUPPORTED|0xbeee),
 #if RETVAL_INJECTED
 	       XLAT_ARGS(IORING_OP_EPOLL_CTL), probe->ops_len, probe->resv2[1],
-	       XLAT_ARGS(IORING_OP_NOP), XLAT_ARGS(IORING_OP_MSG_RING),
+	       XLAT_ARGS(IORING_OP_NOP), XLAT_ARGS(IORING_OP_URING_CMD),
 	       XLAT_ARGS(IO_URING_OP_SUPPORTED),
 	       XLAT_ARGS(IO_URING_OP_SUPPORTED|0xbeee),
 #endif
 	       errstr);
 
-	probe->last_op = 41;
+	probe->last_op = 47;
 	probe->resv2[1] = 0;
 	fill_memory_ex(probe->ops, sizeof(probe->ops[0]) * (DEFAULT_STRLEN + 1),
 		    0x40, 0x80);
@@ -401,7 +394,7 @@ main(void)
 	printf("io_uring_register(%u<%s>, " XLAT_FMT,
 	       fd_null, path_null, XLAT_ARGS(IORING_REGISTER_PROBE));
 	for (size_t c = 0; c < 1 + RETVAL_INJECTED; c++) {
-		printf("%s{last_op=41" NRAW(" /* IORING_OP_??? */")
+		printf("%s{last_op=47" NRAW(" /* IORING_OP_??? */")
 		       ", ops_len=%hhu, ops=[",
 		       c ? " => " : ", ", probe->ops_len);
 		for (size_t i = 0; i < DEFAULT_STRLEN; i++) {
@@ -463,18 +456,18 @@ main(void)
 		{ ARG_STR(IORING_RESTRICTION_REGISTER_OP), true,
 		  "register_op=", ARG_STR(IORING_REGISTER_BUFFERS), true },
 		{ ARG_STR(IORING_RESTRICTION_REGISTER_OP), true,
-		  "register_op=", ARG_STR(IORING_UNREGISTER_RING_FDS),
+		  "register_op=", ARG_STR(IORING_UNREGISTER_PBUF_RING),
 		  true },
 		{ ARG_STR(IORING_RESTRICTION_REGISTER_OP), true,
-		  "register_op=", 22, " /* IORING_REGISTER_??? */", false },
+		  "register_op=", 24, " /* IORING_REGISTER_??? */", false },
 		{ ARG_STR(IORING_RESTRICTION_REGISTER_OP), true,
 		  "register_op=", 255, " /* IORING_REGISTER_??? */", false },
 		{ ARG_STR(IORING_RESTRICTION_SQE_OP), true,
 		  "sqe_op=", ARG_STR(IORING_OP_NOP), true },
 		{ ARG_STR(IORING_RESTRICTION_SQE_OP), true,
-		  "sqe_op=", ARG_STR(IORING_OP_MSG_RING), true },
+		  "sqe_op=", ARG_STR(IORING_OP_URING_CMD), true },
 		{ ARG_STR(IORING_RESTRICTION_SQE_OP), true,
-		  "sqe_op=", 41, " /* IORING_OP_??? */", false },
+		  "sqe_op=", 47, " /* IORING_OP_??? */", false },
 		{ ARG_STR(IORING_RESTRICTION_SQE_OP), true,
 		  "sqe_op=", 255, " /* IORING_OP_??? */", false },
 		{ ARG_STR(IORING_RESTRICTION_SQE_FLAGS_ALLOWED), true,
@@ -652,6 +645,13 @@ main(void)
 		{ 13, "IORING_REGISTER_FILES2" },
 		{ 15, "IORING_REGISTER_BUFFERS2" },
 	};
+	static const struct strval32 rsrc_flags[] = {
+		{ ARG_STR(0) },
+		{ ARG_XLAT_KNOWN(0x1, "IORING_RSRC_REGISTER_SPARSE") },
+		{ ARG_XLAT_UNKNOWN(0x2, "IORING_RSRC_REGISTER_???") },
+		{ ARG_XLAT_KNOWN(0xbadc0ded,
+				 "IORING_RSRC_REGISTER_SPARSE|0xbadc0dec") },
+	};
 	static const uint64_t tags[] = { 0x1337, 1, 0xdead, 0xfacefeed,
 					 0xbadc0deddadfacedULL, 0 };
 	const uint64_t *arg_tags = tail_memdup(tags, sizeof(tags));
@@ -709,7 +709,7 @@ main(void)
 			rsrc_reg->nr += !!(j & 16);
 			rsrc_reg->nr &= ~-!(j & 32);
 
-			rsrc_reg->RESV = j & 64 ? 0xbadc0ded : 0;
+			rsrc_reg->flags = ARR_ITEM(rsrc_flags, j >> 6).val;
 			rsrc_reg->resv2 = j & 128 ? 0xfacecafebeeffeedULL : 0;
 
 			memcpy(big_rsrc_reg, rsrc_reg, sizeof(*rsrc_reg));
@@ -722,14 +722,14 @@ main(void)
 						      sizeof(*rsrc_reg)
 						      + (k / 2) * 8);
 				printf("io_uring_register(%u<%s>, " XLAT_FMT
-				       ", {nr=%zu%s%s",
+				       ", {nr=%zu, flags=%s%s",
 				       fd_null, path_null,
 				       XLAT_SEL(rsrc_reg_ops[i].op,
 						rsrc_reg_ops[i].str),
 				       j & 32 ? (i ? ARRAY_SIZE(iov)
 						   : ARRAY_SIZE(fds))
 						+ !!(j & 16) : 0,
-				       j & 64 ? ", resv=0xbadc0ded" : "",
+				       ARR_ITEM(rsrc_flags, j >> 6).str,
 				       j & 128 ? ", resv2=0xfacecafebeeffeed"
 					       : "");
 				print_rsrc_data(arg_iov, iov, arg_fds, fds,
@@ -1067,6 +1067,67 @@ main(void)
 		}
 	}
 
+
+	/* IORING_REGISTER_PBUF_RING, IORING_UNREGISTER_PBUF_RING */
+	static const struct {
+		unsigned int op;
+		const char *str;
+	} buf_reg_ops[] = {
+		{ 22, "IORING_REGISTER_PBUF_RING" },
+		{ 23, "IORING_UNREGISTER_PBUF_RING" },
+	};
+	TAIL_ALLOC_OBJECT_VAR_PTR(struct io_uring_buf_reg, buf_reg);
+
+	for (size_t i = 0; i < ARRAY_SIZE(buf_reg_ops); i++) {
+		sys_io_uring_register(fd_null, buf_reg_ops[i].op, 0,
+				      0xdeadbeef);
+		printf("io_uring_register(%u<%s>, " XLAT_FMT ", NULL, %u)"
+		       " = %s\n",
+		       fd_null, path_null,
+		       XLAT_SEL(buf_reg_ops[i].op, buf_reg_ops[i].str),
+		       0xdeadbeef, errstr);
+
+		sys_io_uring_register(fd_null, buf_reg_ops[i].op,
+				      buf_reg + 1, 0);
+		printf("io_uring_register(%u<%s>, " XLAT_FMT ", %p, 0) = %s\n",
+		       fd_null, path_null,
+		       XLAT_SEL(buf_reg_ops[i].op, buf_reg_ops[i].str),
+		       buf_reg + 1, errstr);
+
+		for (size_t j = 0; j < 256; j++) {
+			memset(buf_reg, 0, sizeof(*buf_reg));
+			buf_reg->ring_addr = j & 2 ? (uintptr_t) buf_reg : 0;
+			buf_reg->ring_entries = j & 4 ? 3141592653 : 0;
+			buf_reg->bgid = j & 8 ? 42069 : 0;
+			buf_reg->pad = j & 16 ? 31337 : 0;
+			buf_reg->resv[0] = j &  32 ? 0xbadc0deddeadfaceULL : 0;
+			buf_reg->resv[1] = j &  64 ? 0xdecaffedbeefdeadULL : 0;
+			buf_reg->resv[2] = j & 128 ? 0xbadc0dedfacefeedULL : 0;
+
+			sys_io_uring_register(fd_null, buf_reg_ops[i].op,
+					      buf_reg, 0x42);
+			printf("io_uring_register(%u<%s>, " XLAT_FMT
+			       ", {ring_addr=",
+			       fd_null, path_null,
+			       XLAT_SEL(buf_reg_ops[i].op, buf_reg_ops[i].str));
+			if (j & 2)
+				printf("%p", buf_reg);
+			else
+				printf("NULL");
+			printf(", ring_entries=%s, bgid=%s%s",
+			       j & 4 ? "3141592653" : "0",
+			       j & 8 ? "42069" : "0",
+			       j & 16 ? ", pad=0x7a69" : "");
+			if (j & 0xe0) {
+				printf(", resv=[%s, %s, %s]",
+				       j &  32 ? "0xbadc0deddeadface" : "0",
+				       j &  64 ? "0xdecaffedbeefdead" : "0",
+				       j & 128 ? "0xbadc0dedfacefeed" : "0");
+			}
+			printf("}, 66) = %s\n", errstr);
+		}
+	}
+
 	puts("+++ exited with 0 +++");
 	return 0;
 }
