@@ -3,53 +3,31 @@
  *
  * Copyright (c) 2016 Mikulas Patocka <mpatocka@redhat.com>
  * Copyright (c) 2016 Eugene Syromyatnikov <evgsyr@gmail.com>
+ * Copyright (c) 2016-2021 The strace developers.
  * All rights reserved.
  *
- * Redistribution and use in source and binary forms, with or without
- * modification, are permitted provided that the following conditions
- * are met:
- * 1. Redistributions of source code must retain the above copyright
- *    notice, this list of conditions and the following disclaimer.
- * 2. Redistributions in binary form must reproduce the above copyright
- *    notice, this list of conditions and the following disclaimer in the
- *    documentation and/or other materials provided with the distribution.
- * 3. The name of the author may not be used to endorse or promote products
- *    derived from this software without specific prior written permission.
- *
- * THIS SOFTWARE IS PROVIDED BY THE AUTHOR ``AS IS'' AND ANY EXPRESS OR
- * IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED WARRANTIES
- * OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED.
- * IN NO EVENT SHALL THE AUTHOR BE LIABLE FOR ANY DIRECT, INDIRECT,
- * INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT
- * NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
- * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
- * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
- * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
- * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
+ * SPDX-License-Identifier: GPL-2.0-or-later
  */
 
 #include "tests.h"
+#include <errno.h>
+#include <inttypes.h>
+#include <stdio.h>
+#include <stddef.h>
+#include <string.h>
+#include <sys/ioctl.h>
+#include <linux/ioctl.h>
+#include <linux/dm-ioctl.h>
 
-#ifdef HAVE_LINUX_DM_IOCTL_H
+#ifndef VERBOSE
+# define VERBOSE 0
+#endif
 
-# include <errno.h>
-# include <inttypes.h>
-# include <stdbool.h>
-# include <stdio.h>
-# include <stddef.h>
-# include <string.h>
-# include <sys/ioctl.h>
-# include <linux/dm-ioctl.h>
+#define STR32 "AbCdEfGhIjKlMnOpQrStUvWxYz012345"
 
-# ifndef VERBOSE
-#  define VERBOSE 0
-# endif
-
-# define STR32 "AbCdEfGhIjKlMnOpQrStUvWxYz012345"
-
-# define ALIGNED_SIZE(s_, t_) \
+#define ALIGNED_SIZE(s_, t_) \
 	(((s_) + (ALIGNOF(t_) - 1UL)) & ~(ALIGNOF(t_) - 1UL))
-# define ALIGNED_OFFSET(t_, m_) \
+#define ALIGNED_OFFSET(t_, m_) \
 	ALIGNED_SIZE(offsetof(t_, m_), t_)
 
 static const char str129[] = STR32 STR32 STR32 STR32 "6";
@@ -73,9 +51,8 @@ static struct s {
 		} ts;
 		struct {
 			struct dm_target_msg target_msg;
-			char target_string[256];
 		} tm;
-		char string[256];
+		char string[256 + sizeof(struct dm_target_msg)];
 	} u;
 } s;
 
@@ -137,14 +114,14 @@ init_dm_target_spec(struct dm_target_spec *ptr, uint32_t id)
 	ptr->length       = dts_length_base + dts_length_step * id;
 	ptr->status       = dts_status_base + dts_status_step * id;
 
-	strncpy(ptr->target_type, str129 +
+	memcpy(ptr->target_type, str129 +
 		id % (sizeof(str129) - sizeof(ptr->target_type)),
 		id % (sizeof(ptr->target_type) + 1));
 	if (id % (sizeof(ptr->target_type) + 1) < sizeof(ptr->target_type))
 		ptr->target_type[id % (sizeof(ptr->target_type) + 1)] = '\0';
 }
 
-# if VERBOSE
+#if VERBOSE
 static void
 print_dm_target_spec(struct dm_target_spec *ptr, uint32_t id)
 {
@@ -155,11 +132,17 @@ print_dm_target_spec(struct dm_target_spec *ptr, uint32_t id)
 	       (int) (id % (sizeof(ptr->target_type) + 1)),
 	       str129 + id % (sizeof(str129) - sizeof(ptr->target_type)));
 }
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 
 int
 main(void)
 {
+	static kernel_ulong_t dummy_dm_ioctl1 =
+		_IOC(_IOC_READ, DM_IOCTL, 0, 0x1fff);
+	static kernel_ulong_t dummy_dm_ioctl2 =
+		_IOC(_IOC_READ|_IOC_WRITE, DM_IOCTL, 0xed, 0);
+	static kernel_ulong_t dummy_dm_arg =
+		(kernel_ulong_t) 0xbadc0dedda7a1057ULL;
 	/* We can't check these properly for now */
 	static struct args dummy_check_cmds_nodev[] = {
 		{ ARG_STR(DM_REMOVE_ALL),    false },
@@ -174,6 +157,7 @@ main(void)
 		{ ARG_STR(DM_TABLE_CLEAR),   false },
 		{ ARG_STR(DM_TABLE_DEPS),    true  },
 		{ ARG_STR(DM_TABLE_STATUS),  true  },
+		{ ARG_STR(DM_DEV_ARM_POLL),  false },
 	};
 
 	struct dm_ioctl *unaligned_dm_arg =
@@ -191,15 +175,24 @@ main(void)
 
 	long rc;
 	const char *errstr;
-	unsigned int i;
 
 
 	/* Incorrect operation */
 	ioctl(-1, _IOW(DM_IOCTL, 0xde, int), dm_arg);
-	printf("ioctl(-1, _IOC(_IOC_WRITE, %#04x, 0xde, %#04zx), %p) = "
+	printf("ioctl(-1, _IOC(_IOC_WRITE, %#x, 0xde, %#zx), %p) = "
 	       "-1 EBADF (%m)\n",
 	       DM_IOCTL, sizeof(int), dm_arg);
 
+	ioctl(-1, dummy_dm_ioctl1, 0);
+	printf("ioctl(-1, _IOC(_IOC_READ, %#x, 0, %#x), 0) = -1 EBADF (%m)\n",
+	       DM_IOCTL, (unsigned int) _IOC_SIZE(dummy_dm_ioctl1));
+
+	ioctl(-1, dummy_dm_ioctl2, dummy_dm_arg);
+	printf("ioctl(-1, _IOC(_IOC_READ|_IOC_WRITE, %#x, %#x, 0), %#lx) = "
+	       "-1 EBADF (%m)\n",
+	       DM_IOCTL, (unsigned int) _IOC_NR(dummy_dm_ioctl2),
+	       (unsigned long) dummy_dm_arg);
+
 
 	/* DM_VERSION */
 	/* Incorrect pointer */
@@ -217,23 +210,23 @@ main(void)
 	dm_arg->version[1] = 0xbadc0dee;
 	dm_arg->version[2] = 0xbadc0def;
 	ioctl(-1, DM_VERSION, dm_arg);
-	printf("ioctl(-1, DM_VERSION, {version=%u.%u.%u, "
-	       "/* Unsupported device mapper ABI version */ ...}) = "
+	printf("ioctl(-1, DM_VERSION, {version=[%u, %u, %u]"
+	       " /* unsupported device mapper ABI version */}) = "
 	       "-1 EBADF (%m)\n", 0xbadc0ded, 0xbadc0dee, 0xbadc0def);
 
 	/* Incorrect data_size */
 	init_s(dm_arg, 14, 64);
 	ioctl(-1, DM_VERSION, dm_arg);
-	printf("ioctl(-1, DM_VERSION, {version=4.1.2, data_size=14, "
-	       "/* Incorrect data_size */ ...}) = -1 EBADF (%m)\n");
+	printf("ioctl(-1, DM_VERSION, {version=[4, 1, 2], data_size=14"
+	       " /* data_size too small */}) = -1 EBADF (%m)\n");
 
 	/* Unterminated name/uuid */
 	init_s(dm_arg, min_sizeof_dm_ioctl, 0);
-	strncpy(dm_arg->name, str129, sizeof(dm_arg->name));
-	strncpy(dm_arg->uuid, str129, sizeof(dm_arg->uuid));
+	memcpy(dm_arg->name, str129, sizeof(dm_arg->name));
+	memcpy(dm_arg->uuid, str129, sizeof(dm_arg->uuid));
 	ioctl(-1, DM_VERSION, dm_arg);
-	printf("ioctl(-1, DM_VERSION, {version=4.1.2, data_size=%zu, "
-	       "dev=makedev(18, 52), name=\"%.127s\", uuid=\"%.128s\", "
+	printf("ioctl(-1, DM_VERSION, {version=[4, 1, 2], data_size=%zu, "
+	       "dev=makedev(0x12, 0x34), name=\"%.127s\"..., uuid=\"%.128s\"..., "
 	       "flags=0}) = -1 EBADF (%m)\n",
 	       min_sizeof_dm_ioctl, str129, str129);
 
@@ -241,8 +234,8 @@ main(void)
 	init_s(dm_arg, min_sizeof_dm_ioctl, 0);
 	ioctl(-1, DM_VERSION, dm_arg);
 	printf("ioctl(-1, DM_VERSION, "
-	       "{version=4.1.2, data_size=%zu, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", flags=0}) = "
+	       "{version=[4, 1, 2], data_size=%zu, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", flags=0}) = "
 	       "-1 EBADF (%m)\n", min_sizeof_dm_ioctl);
 
 	/* Zero dev, name, uuid */
@@ -253,7 +246,7 @@ main(void)
 	dm_arg->uuid[0] = '\0';
 	ioctl(-1, DM_VERSION, dm_arg);
 	printf("ioctl(-1, DM_VERSION, "
-	       "{version=4.1.2, data_size=%u, flags=0}) = "
+	       "{version=[4, 1, 2], data_size=%u, flags=0}) = "
 	       "-1 EBADF (%m)\n", 0xfacefeed);
 
 	/* Flag */
@@ -261,8 +254,8 @@ main(void)
 	dm_arg->flags = 0xffffffff;
 	ioctl(-1, DM_VERSION, dm_arg);
 	printf("ioctl(-1, DM_VERSION, "
-	       "{version=4.1.2, data_size=%zu, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", flags="
+	       "{version=[4, 1, 2], data_size=%zu, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", flags="
 	       "DM_READONLY_FLAG|DM_SUSPEND_FLAG|DM_EXISTS_FLAG|"
 	       "DM_PERSISTENT_DEV_FLAG|DM_STATUS_TABLE_FLAG|"
 	       "DM_ACTIVE_PRESENT_FLAG|DM_INACTIVE_PRESENT_FLAG|"
@@ -270,25 +263,25 @@ main(void)
 	       "DM_NOFLUSH_FLAG|DM_QUERY_INACTIVE_TABLE_FLAG|"
 	       "DM_UEVENT_GENERATED_FLAG|DM_UUID_FLAG|DM_SECURE_DATA_FLAG|"
 	       "DM_DATA_OUT_FLAG|DM_DEFERRED_REMOVE|DM_INTERNAL_SUSPEND_FLAG|"
-	       "0xfff80080}) = -1 EBADF (%m)\n",
+	       "DM_IMA_MEASUREMENT_FLAG|0xfff00080}) = -1 EBADF (%m)\n",
 	       min_sizeof_dm_ioctl);
 
 	/* Normal call */
 	init_s(&s.ioc, sizeof(s.ioc), 0);
 	ioctl(-1, DM_VERSION, &s);
 	printf("ioctl(-1, DM_VERSION, "
-	       "{version=4.1.2, data_size=%zu, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", flags=0}) = "
+	       "{version=[4, 1, 2], data_size=%zu, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", flags=0}) = "
 	       "-1 EBADF (%m)\n", sizeof(s.ioc));
 
 
 	/* DM_REMOVE_ALL */
 	/* DM_LIST_DEVICES */
 	/* DM_LIST_VERSIONS */
-	for (i = 0; i < ARRAY_SIZE(dummy_check_cmds_nodev); i++) {
+	for (unsigned int i = 0; i < ARRAY_SIZE(dummy_check_cmds_nodev); ++i) {
 		init_s(dm_arg, min_sizeof_dm_ioctl, 0);
 		ioctl(-1, dummy_check_cmds_nodev[i].arg, dm_arg);
-		printf("ioctl(-1, %s, {version=4.1.2, data_size=%zu%s, "
+		printf("ioctl(-1, %s, {version=[4, 1, 2], data_size=%zu%s, "
 		       "flags=0}) = -1 EBADF (%m)\n",
 		       dummy_check_cmds_nodev[i].str,
 		       min_sizeof_dm_ioctl,
@@ -304,11 +297,11 @@ main(void)
 	/* DM_TABLE_CLEAR */
 	/* DM_TABLE_DEPS */
 	/* DM_TABLE_STATUS */
-	for (i = 0; i < ARRAY_SIZE(dummy_check_cmds); i++) {
+	for (unsigned int i = 0; i < ARRAY_SIZE(dummy_check_cmds); ++i) {
 		init_s(dm_arg, min_sizeof_dm_ioctl, 0);
 		ioctl(-1, dummy_check_cmds[i].arg, dm_arg);
-		printf("ioctl(-1, %s, {version=4.1.2, data_size=%zu%s, "
-		       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\"%s, "
+		printf("ioctl(-1, %s, {version=[4, 1, 2], data_size=%zu%s, "
+		       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\"%s, "
 		       "flags=0}) = -1 EBADF (%m)\n",
 		       dummy_check_cmds[i].str, min_sizeof_dm_ioctl,
 		       dummy_check_cmds[i].has_params ? ", data_start=0" : "",
@@ -322,15 +315,15 @@ main(void)
 	s.ioc.event_nr = 0xbadc0ded;
 	ioctl(-1, DM_DEV_SUSPEND, &s);
 	printf("ioctl(-1, DM_DEV_SUSPEND, "
-	       "{version=4.1.2, data_size=%zu, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", "
+	       "{version=[4, 1, 2], data_size=%zu, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", "
 	       "flags=DM_SUSPEND_FLAG}) = -1 EBADF (%m)\n", sizeof(s.ioc));
 
 	init_s(&s.ioc, sizeof(s.ioc), 0);
 	s.ioc.event_nr = 0xbadc0ded;
 	ioctl(-1, DM_DEV_SUSPEND, &s);
 	printf("ioctl(-1, DM_DEV_SUSPEND, "
-	       "{version=4.1.2, data_size=%zu, dev=makedev(18, 52), "
+	       "{version=[4, 1, 2], data_size=%zu, dev=makedev(0x12, 0x34), "
 	       "name=\"nnn\", uuid=\"uuu\", event_nr=3134983661, "
 	       "flags=0}) = -1 EBADF (%m)\n", sizeof(s.ioc));
 
@@ -346,15 +339,15 @@ main(void)
 	strcpy(s.u.ts.target_params, "tparams");
 	ioctl(-1, DM_TABLE_LOAD, &s);
 	printf("ioctl(-1, DM_TABLE_LOAD, "
-	       "{version=4.1.2, data_size=%u, data_start=%u, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", "
+	       "{version=[4, 1, 2], data_size=%u, data_start=%u, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", "
 	       "target_count=1, flags=0, "
-# if VERBOSE
+#if VERBOSE
 	       "{sector_start=16, length=32, target_type=\"tgt\", "
 	       "string=\"tparams\"}"
-# else /* !VERBOSE */
+#else /* !VERBOSE */
 	       "..."
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 	       "}) = -1 EBADF (%m)\n", s.ioc.data_size, s.ioc.data_start);
 
 	/* No targets */
@@ -363,8 +356,8 @@ main(void)
 	dm_arg->target_count = 0;
 	ioctl(-1, DM_TABLE_LOAD, dm_arg);
 	printf("ioctl(-1, DM_TABLE_LOAD, "
-	       "{version=4.1.2, data_size=%zu, data_start=%zu, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", "
+	       "{version=[4, 1, 2], data_size=%zu, data_start=%zu, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", "
 	       "target_count=0, flags=0}) = -1 EBADF (%m)\n",
 	       sizeof(*dm_arg), min_sizeof_dm_ioctl);
 
@@ -374,14 +367,14 @@ main(void)
 	dm_arg->target_count = 1234;
 	ioctl(-1, DM_TABLE_LOAD, dm_arg);
 	printf("ioctl(-1, DM_TABLE_LOAD, "
-	       "{version=4.1.2, data_size=%zu, data_start=%u, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", "
+	       "{version=[4, 1, 2], data_size=%zu, data_start=%u, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", "
 	       "target_count=1234, flags=0, "
-# if VERBOSE
-	       "/* misplaced struct dm_target_spec */ ..."
-# else /* !VERBOSE */
+#if VERBOSE
+	       "??? /* misplaced struct dm_target_spec */"
+#else
 	       "..."
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 	       "}) = -1 EBADF (%m)\n", sizeof(*dm_arg), 0xfffffff8);
 
 	/* Inaccessible pointer */
@@ -391,20 +384,20 @@ main(void)
 	dm_arg_open1->ioc.target_count = 0xdeaddea1;
 	ioctl(-1, DM_TABLE_LOAD, dm_arg_open1);
 	printf("ioctl(-1, DM_TABLE_LOAD, "
-	       "{version=4.1.2, data_size=%zu, data_start=%zu, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", "
+	       "{version=[4, 1, 2], data_size=%zu, data_start=%zu, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", "
 	       "target_count=3735936673, flags=0, "
-# if VERBOSE
+#if VERBOSE
 	       "%p"
-# else /* !VERBOSE */
+#else /* !VERBOSE */
 	       "..."
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 	       "}) = -1 EBADF (%m)\n", sizeof(*dm_arg_open1),
 	       offsetof(struct dm_table_open_test, target1)
-# if VERBOSE
+#if VERBOSE
 	       , (char *) dm_arg_open1 +
 	       offsetof(struct dm_table_open_test, target1)
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 	       );
 
 	/* Inaccessible string */
@@ -419,21 +412,21 @@ main(void)
 	rc = ioctl(-1, DM_TABLE_LOAD, dm_arg_open2);
 	errstr = sprintrc(rc);
 	printf("ioctl(-1, DM_TABLE_LOAD, "
-	       "{version=4.1.2, data_size=%zu, data_start=%zu, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", "
+	       "{version=[4, 1, 2], data_size=%zu, data_start=%zu, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", "
 	       "target_count=2, flags=0, ",
 	       sizeof(*dm_arg_open2),
 	       offsetof(struct dm_table_open_test, target1));
-# if VERBOSE
+#if VERBOSE
 	print_dm_target_spec(&dm_arg_open2->target1, 7);
 	printf("%p}, %p",
 	       (char *) dm_arg_open2 +
 	       offsetof(struct dm_table_open_test, param1),
 	       (char *) dm_arg_open2 +
 	       offsetof(struct dm_table_open_test, target3));
-# else /* !VERBOSE */
+#else /* !VERBOSE */
 	printf("...");
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 	printf("}) = %s\n", errstr);
 
 	/* Incorrect next */
@@ -463,24 +456,24 @@ main(void)
 	rc = ioctl(-1, DM_TABLE_LOAD, dm_arg_open3);
 	errstr = sprintrc(rc);
 	printf("ioctl(-1, DM_TABLE_LOAD, "
-	       "{version=4.1.2, data_size=%zu, data_start=%zu, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", "
+	       "{version=[4, 1, 2], data_size=%zu, data_start=%zu, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", "
 	       "target_count=4, flags=0, ",
 	       offsetof(struct dm_table_open_test, target5),
 	       offsetof(struct dm_table_open_test, target0));
-# if VERBOSE
+#if VERBOSE
 	print_dm_target_spec(&dm_arg_open3->target0, 9);
 	printf("\"\"}, ");
 	print_dm_target_spec(&dm_arg_open3->target1, 15);
 	printf("\"\\377\"}, ");
 	print_dm_target_spec(&dm_arg_open3->target1, 42);
-	printf("\"\\1\\2\"}, /* misplaced struct dm_target_spec */ ...");
-# else /* !VERBOSE */
+	printf("\"\\1\\2\"}, ??? /* misplaced struct dm_target_spec */");
+#else /* !VERBOSE */
 	printf("...");
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 	printf("}) = %s\n", errstr);
 
-	#define FILL_DM_TARGET(id, id_next) \
+#define FILL_DM_TARGET(id, id_next) \
 		do { \
 			init_dm_target_spec(&dm_arg_open3->target##id, id); \
 			dm_arg_open3->target##id.next = \
@@ -488,10 +481,10 @@ main(void)
 					 target##id_next) - \
 				offsetof(struct dm_table_open_test, \
 					 target##id); \
-			strncpy(dm_arg_open3->param##id, str129 + id * 2, id); \
+			memcpy(dm_arg_open3->param##id, str129 + id * 2, id); \
 			dm_arg_open3->param##id[id] = '\0'; \
 		} while (0)
-	#define PRINT_DM_TARGET(id) \
+#define PRINT_DM_TARGET(id) \
 		do { \
 			print_dm_target_spec(&dm_arg_open3->target##id, id); \
 			printf("\"%.*s\"}, ", id, str129 + id * 2); \
@@ -514,12 +507,12 @@ main(void)
 	rc = ioctl(-1, DM_TABLE_LOAD, dm_arg_open3);
 	errstr = sprintrc(rc);
 	printf("ioctl(-1, DM_TABLE_LOAD, "
-	       "{version=4.1.2, data_size=%zu, data_start=%zu, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", "
+	       "{version=[4, 1, 2], data_size=%zu, data_start=%zu, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", "
 	       "target_count=3134983661, flags=0, ",
 	       sizeof(*dm_arg_open3),
 	       offsetof(struct dm_table_open_test, target0));
-# if VERBOSE
+#if VERBOSE
 	PRINT_DM_TARGET(0);
 	PRINT_DM_TARGET(1);
 	PRINT_DM_TARGET(2);
@@ -529,7 +522,7 @@ main(void)
 	PRINT_DM_TARGET(6);
 	PRINT_DM_TARGET(7);
 	PRINT_DM_TARGET(8);
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 	printf("...}) = %s\n", errstr);
 
 
@@ -540,13 +533,13 @@ main(void)
 	       "long target msg");
 	ioctl(-1, DM_TARGET_MSG, &s);
 	printf("ioctl(-1, DM_TARGET_MSG, "
-	       "{version=4.1.2, data_size=%u, data_start=%u, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", flags=0, "
-# if VERBOSE
+	       "{version=[4, 1, 2], data_size=%u, data_start=%u, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", flags=0, "
+#if VERBOSE
 	       "{sector=4660, message=\"long targ\"...}"
-# else /* !VERBOSE */
+#else /* !VERBOSE */
 	       "..."
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 	       "}) = -1 EBADF (%m)\n",
 	       s.ioc.data_size, s.ioc.data_start);
 
@@ -555,13 +548,13 @@ main(void)
 	dm_arg->data_size = sizeof(*dm_arg);
 	ioctl(-1, DM_TARGET_MSG, dm_arg);
 	printf("ioctl(-1, DM_TARGET_MSG, "
-	       "{version=4.1.2, data_size=%zu, data_start=%zu, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", flags=0, "
-# if VERBOSE
-	       "/* misplaced struct dm_target_msg */"
-# else /* !VERBOSE */
+	       "{version=[4, 1, 2], data_size=%zu, data_start=%zu, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", flags=0, "
+#if VERBOSE
+	       "??? /* misplaced struct dm_target_msg */"
+#else /* !VERBOSE */
 	       "..."
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 	       "}) = -1 EBADF (%m)\n",
 	       sizeof(*dm_arg), min_sizeof_dm_ioctl);
 
@@ -570,13 +563,13 @@ main(void)
 	dm_arg->data_size = sizeof(*dm_arg);
 	ioctl(-1, DM_TARGET_MSG, dm_arg);
 	printf("ioctl(-1, DM_TARGET_MSG, "
-	       "{version=4.1.2, data_size=%zu, data_start=%u, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", flags=0, "
-# if VERBOSE
-	       "/* misplaced struct dm_target_msg */"
-# else /* !VERBOSE */
+	       "{version=[4, 1, 2], data_size=%zu, data_start=%u, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", flags=0, "
+#if VERBOSE
+	       "??? /* misplaced struct dm_target_msg */"
+#else /* !VERBOSE */
 	       "..."
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 	       "}) = -1 EBADF (%m)\n",
 	       sizeof(*dm_arg), 0xffffffff);
 
@@ -586,19 +579,19 @@ main(void)
 	dm_arg->data_start = sizeof(*dm_arg);
 	ioctl(-1, DM_TARGET_MSG, dm_arg);
 	printf("ioctl(-1, DM_TARGET_MSG, "
-	       "{version=4.1.2, data_size=%zu, data_start=%zu, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", flags=0, "
-# if VERBOSE
+	       "{version=[4, 1, 2], data_size=%zu, data_start=%zu, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", flags=0, "
+#if VERBOSE
 	       "%p"
-# else /* !VERBOSE */
+#else /* !VERBOSE */
 	       "..."
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 	       "}) = -1 EBADF (%m)\n",
 	       sizeof(*dm_arg) + sizeof(struct dm_target_msg),
 	       sizeof(*dm_arg)
-# if VERBOSE
+#if VERBOSE
 	       , (char *) dm_arg + sizeof(*dm_arg)
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 	       );
 
 	/* Inaccessible string */
@@ -609,18 +602,18 @@ main(void)
 	rc = ioctl(-1, DM_TARGET_MSG, dm_arg_msg);
 	errstr = sprintrc(rc);
 	printf("ioctl(-1, DM_TARGET_MSG, "
-	       "{version=4.1.2, data_size=%zu, data_start=%zu, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", flags=0, ",
+	       "{version=[4, 1, 2], data_size=%zu, data_start=%zu, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", flags=0, ",
 	       sizeof(*dm_arg_msg) + 1,
 	       offsetof(struct dm_target_msg_test, msg));
-# if VERBOSE
+#if VERBOSE
 	printf("{sector=%" PRI__u64 ", message=%p}",
 	       (__u64) 0xdeadbeeffacef157ULL,
 	       (char *) dm_arg_msg +
 	       offsetof(struct dm_target_msg_test, msg.message));
-# else /* !VERBOSE */
+#else /* !VERBOSE */
 	printf("...");
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 	printf("}) = %s\n", errstr);
 
 	/* Zero-sied string */
@@ -630,15 +623,15 @@ main(void)
 	rc = ioctl(-1, DM_TARGET_MSG, dm_arg_msg);
 	errstr = sprintrc(rc);
 	printf("ioctl(-1, DM_TARGET_MSG, "
-	       "{version=4.1.2, data_size=%zu, data_start=%zu, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", flags=0, ",
+	       "{version=[4, 1, 2], data_size=%zu, data_start=%zu, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", flags=0, ",
 	       sizeof(*dm_arg_msg), offsetof(struct dm_target_msg_test, msg));
-# if VERBOSE
+#if VERBOSE
 	printf("{sector=%" PRI__u64 ", message=\"\"}",
 	       (__u64) 0xdeadbeeffacef157ULL);
-# else /* !VERBOSE */
+#else /* !VERBOSE */
 	printf("...");
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 	printf("}) = %s\n", errstr);
 
 
@@ -647,13 +640,13 @@ main(void)
 	strcpy(s.u.string, "10 20 30 40");
 	ioctl(-1, DM_DEV_SET_GEOMETRY, &s);
 	printf("ioctl(-1, DM_DEV_SET_GEOMETRY, "
-	       "{version=4.1.2, data_size=%u, data_start=%u, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", flags=0, "
-# if VERBOSE
+	       "{version=[4, 1, 2], data_size=%u, data_start=%u, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", flags=0, "
+#if VERBOSE
 	       "string=\"10 20 30 \"..."
-# else /* !VERBOSE */
+#else /* !VERBOSE */
 	       "..."
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 	       "}) = -1 EBADF (%m)\n",
 	       s.ioc.data_size, s.ioc.data_start);
 
@@ -665,19 +658,19 @@ main(void)
 	memcpy(unaligned_dm_arg, dm_arg, offsetof(struct dm_ioctl, data));
 	ioctl(-1, DM_DEV_RENAME, unaligned_dm_arg);
 	printf("ioctl(-1, DM_DEV_RENAME, "
-	       "{version=4.1.2, data_size=%zu, data_start=%zu, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", event_nr=0, "
+	       "{version=[4, 1, 2], data_size=%zu, data_start=%zu, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", event_nr=0, "
 	       "flags=0, "
-# if VERBOSE
+#if VERBOSE
 	       "string=%p"
-# else /* !VERBOSE */
+#else /* !VERBOSE */
 	       "..."
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 	       "}) = -1 EBADF (%m)\n",
 	       sizeof(*unaligned_dm_arg), min_sizeof_dm_ioctl
-# if VERBOSE
+#if VERBOSE
 	       , (char *) unaligned_dm_arg + min_sizeof_dm_ioctl
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 	       );
 
 	/* Incorrect data_start data */
@@ -685,14 +678,14 @@ main(void)
 	s.ioc.data_start = 0xdeadbeef;
 	ioctl(-1, DM_DEV_RENAME, &s);
 	printf("ioctl(-1, DM_DEV_RENAME, "
-	       "{version=4.1.2, data_size=%u, data_start=3735928559, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", event_nr=0, "
+	       "{version=[4, 1, 2], data_size=%u, data_start=3735928559, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", event_nr=0, "
 	       "flags=0, "
-# if VERBOSE
-	       "/* misplaced string */"
-# else /* !VERBOSE */
+#if VERBOSE
+	       "??? /* misplaced string */"
+#else /* !VERBOSE */
 	       "..."
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 	       "}) = -1 EBADF (%m)\n",
 	       s.ioc.data_size);
 
@@ -702,14 +695,14 @@ main(void)
 	s.ioc.data_start = offsetof(struct dm_ioctl, name) + 1;
 	ioctl(-1, DM_DEV_RENAME, &s);
 	printf("ioctl(-1, DM_DEV_RENAME, "
-	       "{version=4.1.2, data_size=%u, data_start=%zu, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", event_nr=0, "
+	       "{version=[4, 1, 2], data_size=%u, data_start=%zu, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", event_nr=0, "
 	       "flags=0, "
-# if VERBOSE
+#if VERBOSE
 	       "string=\"nn\""
-# else /* !VERBOSE */
+#else /* !VERBOSE */
 	       "..."
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 	       "}) = -1 EBADF (%m)\n",
 	       s.ioc.data_size,
 	       offsetof(struct dm_ioctl, name) + 1);
@@ -719,14 +712,14 @@ main(void)
 	strcpy(s.u.string, "new long name");
 	ioctl(-1, DM_DEV_RENAME, &s);
 	printf("ioctl(-1, DM_DEV_RENAME, "
-	       "{version=4.1.2, data_size=%u, data_start=%u, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", event_nr=0, "
+	       "{version=[4, 1, 2], data_size=%u, data_start=%u, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", event_nr=0, "
 	       "flags=0, "
-# if VERBOSE
+#if VERBOSE
 	       "string=\"new long \"..."
-# else /* !VERBOSE */
+#else /* !VERBOSE */
 	       "..."
-# endif /* VERBOSE */
+#endif /* VERBOSE */
 	       "}) = -1 EBADF (%m)\n",
 	       s.ioc.data_size, s.ioc.data_start);
 
@@ -736,22 +729,18 @@ main(void)
 	s.ioc.target_count = -1U;
 	ioctl(-1, DM_TABLE_LOAD, &s);
 	printf("ioctl(-1, DM_TABLE_LOAD, "
-	       "{version=4.1.2, data_size=%u, data_start=%u, "
-	       "dev=makedev(18, 52), name=\"nnn\", uuid=\"uuu\", "
+	       "{version=[4, 1, 2], data_size=%u, data_start=%u, "
+	       "dev=makedev(0x12, 0x34), name=\"nnn\", uuid=\"uuu\", "
 	       "target_count=4294967295, flags=0, "
-# if VERBOSE
-	       "{sector_start=0, length=0, target_type=\"\", string=\"\"}, "
-	       "/* misplaced struct dm_target_spec */ "
-# endif /* VERBOSE */
-	       "...}) = -1 EBADF (%m)\n",
+#if VERBOSE
+	       "{sector_start=0, length=0, target_type=\"\", string=\"\"}"
+	       ", ??? /* misplaced struct dm_target_spec */"
+#else
+	       "..."
+#endif /* VERBOSE */
+	       "}) = -1 EBADF (%m)\n",
 	       s.ioc.data_size, s.ioc.data_start);
 
 	puts("+++ exited with 0 +++");
 	return 0;
 }
-
-#else /* !HAVE_LINUX_DM_IOCTL_H */
-
-SKIP_MAIN_UNDEFINED("HAVE_LINUX_DM_IOCTL_H")
-
-#endif /* HAVE_LINUX_DM_IOCTL_H */
