@@ -5,7 +5,7 @@
  * Copyright (c) 2020-2021 The strace developers.
  * All rights reserved.
  *
- * SPDX-License-Identifier: LGPL-2.1-or-later
+ * SPDX-License-Identifier: GPL-2.0-or-later
  */
 
 #include "tests.h"
@@ -23,15 +23,7 @@
 # include <sys/wait.h>
 # include <unistd.h>
 # include <linux/sched.h>
-# include "nsfs.h"
-
-# ifndef CLONE_NEWUSER
-#  define CLONE_NEWUSER 0x10000000
-# endif
-
-# ifndef CLONE_NEWPID
-#  define CLONE_NEWPID 0x20000000
-# endif
+# include <linux/nsfs.h>
 
 static int
 fork_chain(int depth)
